@@ -6,12 +6,6 @@ set -o pipefail
 CI_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
 source "${CI_DIR}/common/build.sh"
 
-# Enable ipv6 on Travis. ref: a39c8b7ce30d
-if test -n "${TRAVIS_OS_NAME}" && ! test "${TRAVIS_OS_NAME}" = osx ; then
-  echo "before_script.sh: enable ipv6"
-  sudo sysctl -w net.ipv6.conf.lo.disable_ipv6=0
-fi
-
 # Test some of the configuration variables.
 if [[ -n "${GCOV}" ]] && [[ ! $(type -P "${GCOV}") ]]; then
   echo "\$GCOV: '${GCOV}' is not executable."
@@ -27,13 +21,6 @@ ccache -s
 # Reset ccache stats for real results in before_cache.
 ccache --zero-stats
 
-if [[ "${TRAVIS_OS_NAME}" == osx ]]; then
-  # Adds user to a dummy group.
-  # That allows to test changing the group of the file by `os_fchown`.
-  sudo dscl . -create /Groups/chown_test
-  sudo dscl . -append /Groups/chown_test GroupMembership "${USER}"
-fi
-
 # Compile dependencies.
 build_deps
 
