@@ -17,7 +17,8 @@ function M.check()
   local log_path = vim.lsp.get_log_path()
   report_info(string.format("Log path: %s", log_path))
 
-  local log_size = vim.loop.fs_stat(log_path).size
+  local log_file = vim.loop.fs_stat(log_path)
+  local log_size = log_file and log_file.size or 0
 
   local report_fn = (log_size / 1000000 > 100 and report_warn or report_info)
   report_fn(string.format("Log size: %d KB", log_size / 1000 ))
