@@ -36,7 +36,7 @@ usage() {
   echo "                       can be a Vim version (8.0.xxx) or a Git hash."
   echo "    -P {vim-revision}  Download, generate and apply a Vim patch."
   echo "    -g {vim-revision}  Download a Vim patch."
-  echo "    -s                 Create a vim-patch pull request."
+  echo "    -s [pr args]       Create a vim-patch pull request."
   echo "    -r {pr-number}     Review a vim-patch pull request."
   echo "    -V                 Clone the Vim source code to \$VIM_SOURCE_DIR."
   echo
@@ -191,8 +191,8 @@ preprocess_patch() {
 ' +w +q "$file"
 
   # Remove *.proto, Make*, INSTALL*, gui_*, beval.*, some if_*, gvim, libvterm, tee, VisVim, xpm, xxd
-  local na_src='auto\|configure.*\|GvimExt\|libvterm\|proto\|tee\|VisVim\|xpm\|xxd\|Make*\|INSTALL*\|beval.*\|gui_*\|if_lua\|if_mzsch\|if_olepp\|if_ole\|if_perl\|if_py\|if_ruby\|if_tcl\|if_xcmdsrv'
-  2>/dev/null $nvim --cmd 'set dir=/tmp' +'g@^diff --git a/src/\S*\<\%(testdir/\)\@<!\%('"${na_src}"'\)@norm! d/\v(^diff)|%$
+  local na_src='auto\|configure.*\|GvimExt\|libvterm\|proto\|tee\|VisVim\|xpm\|xxd\|Make.*\|INSTALL.*\|beval.*\|gui.*\|if_lua\|if_mzsch\|if_olepp\|if_ole\|if_perl\|if_py\|if_ruby\|if_tcl\|if_xcmdsrv'
+  2>/dev/null $nvim --cmd 'set dir=/tmp' +'g@^diff --git a/src/\S*\<\%(testdir/\)\@<!\%('"${na_src}"'\)\>@norm! d/\v(^diff)|%$
 ' +w +q "$file"
 
   # Remove unwanted Vim doc files.
@@ -204,22 +204,21 @@ preprocess_patch() {
   2>/dev/null $nvim --cmd 'set dir=/tmp' +'%s/^@@.*\n.*For Vim version.*Last change.*\n.*For Vim version.*Last change.*//' +w +q "$file"
 
   # Remove gui, option, setup, screen dumps, testdir/Make_*.mak files
-  local na_src_testdir='gen_opt_test.vim\|gui_.*\|Make_amiga.mak\|Make_dos.mak\|Make_ming.mak\|Make_vms.mms\|dumps/.*.dump\|setup_gui.vim'
+  local na_src_testdir='gen_opt_test\.vim\|gui_.*\|Make_amiga\.mak\|Make_dos\.mak\|Make_ming\.mak\|Make_vms\.mms\|dumps/.*\.dump\|setup_gui\.vim'
   2>/dev/null $nvim --cmd 'set dir=/tmp' +'g@^diff --git a/src/testdir/\<\%('"${na_src_testdir}"'\)\>@norm! d/\v(^diff)|%$
 ' +w +q "$file"
 
   # Remove testdir/test_*.vim files
-  local na_src_testdir='balloon.*\|channel.*\|crypt.vim\|gui.*\|job_fails.vim\|json.vim\|mzscheme.vim\|netbeans.*\|paste.vim\|popupwin.*\|restricted.vim\|shortpathname.vim\|tcl.vim\|terminal.*\|xxd.vim'
+  local na_src_testdir='balloon.*\|channel.*\|crypt\.vim\|gui.*\|job_fails\.vim\|json\.vim\|mzscheme\.vim\|netbeans.*\|paste\.vim\|popupwin.*\|restricted\.vim\|shortpathname\.vim\|tcl\.vim\|terminal.*\|xxd\.vim'
   2>/dev/null $nvim --cmd 'set dir=/tmp' +'g@^diff --git a/src/testdir/\<test_\%('"${na_src_testdir}"'\)\>@norm! d/\v(^diff)|%$
 ' +w +q "$file"
 
   # Remove version.c #7555
-  local na_po='version.c'
-  2>/dev/null $nvim --cmd 'set dir=/tmp' +'g@^diff --git a/src/\<\%('${na_po}'\)\>@norm! d/\v(^diff)|%$
+  2>/dev/null $nvim --cmd 'set dir=/tmp' +'g@^diff --git a/src/\<\%(version\.c\)\>@norm! d/\v(^diff)|%$
 ' +w +q "$file"
 
   # Remove some *.po files. #5622
-  local na_po='sjiscorr.c\|ja.sjis.po\|ko.po\|pl.cp1250.po\|pl.po\|ru.cp1251.po\|uk.cp1251.po\|zh_CN.cp936.po\|zh_CN.po\|zh_TW.po'
+  local na_po='sjiscorr\.c\|ja\.sjis\.po\|ko\.po\|pl\.cp1250\.po\|pl\.po\|ru\.cp1251\.po\|uk\.cp1251\.po\|zh_CN\.cp936\.po\|zh_CN\.po\|zh_TW\.po'
   2>/dev/null $nvim --cmd 'set dir=/tmp' +'g@^diff --git a/src/po/\<\%('${na_po}'\)\>@norm! d/\v(^diff)|%$
 ' +w +q "$file"
 
@@ -244,12 +243,16 @@ preprocess_patch() {
   LC_ALL=C sed -e 's/\( [ab]\/src\/nvim\)\/session\(\.[ch]\)/\1\/ex_session\2/g' \
     "$file" > "$file".tmp && mv "$file".tmp "$file"
 
+  # Rename highlight.c to highlight_group.c
+  LC_ALL=C sed -e 's/\( [ab]\/src\/nvim\)\/highlight\(\.[ch]\)/\1\/highlight_group\2/g' \
+    "$file" > "$file".tmp && mv "$file".tmp "$file"
+
   # Rename test_urls.vim to check_urls.vim
-  LC_ALL=C sed -e 's@\( [ab]\)/runtime/doc/test\(_urls.vim\)@\1/scripts/check\2@g' \
+  LC_ALL=C sed -e 's@\( [ab]\)/runtime/doc/test\(_urls\.vim\)@\1/scripts/check\2@g' \
     "$file" > "$file".tmp && mv "$file".tmp "$file"
 
   # Rename path to check_colors.vim
-  LC_ALL=C sed -e 's@\( [ab]/runtime\)/colors/\(tools/check_colors.vim\)@\1/\2@g' \
+  LC_ALL=C sed -e 's@\( [ab]/runtime\)/colors/\(tools/check_colors\.vim\)@\1/\2@g' \
     "$file" > "$file".tmp && mv "$file".tmp "$file"
 }
 
@@ -274,8 +277,6 @@ get_vimpatch() {
   msg_ok "Saved patch to '${NVIM_SOURCE_DIR}/${patch_file}'."
 }
 
-# shellcheck disable=SC2015
-# ^ "Note that A && B || C is not if-then-else."
 stage_patch() {
   get_vimpatch "$1"
   local try_apply="${2:-}"
@@ -290,23 +291,32 @@ stage_patch() {
     echo "  branch; not creating a new branch."
   else
     printf '\nFetching "%s/master".\n' "${nvim_remote}"
-    output="$(git fetch "${nvim_remote}" master 2>&1)" &&
-      msg_ok "${output}" ||
-      (msg_err "${output}"; false)
+    if output="$(git fetch "$nvim_remote" master 2>&1)"; then
+      msg_ok "$output"
+    else
+      msg_err "$output"
+      exit 1
+    fi
 
     local nvim_branch="${BRANCH_PREFIX}${vim_version}"
     echo
     echo "Creating new branch '${nvim_branch}' based on '${nvim_remote}/master'."
     cd "${NVIM_SOURCE_DIR}"
-    output="$(git checkout -b "${nvim_branch}" "${nvim_remote}/master" 2>&1)" &&
-      msg_ok "${output}" ||
-      (msg_err "${output}"; false)
+    if output="$(git checkout -b "$nvim_branch" "$nvim_remote/master" 2>&1)"; then
+      msg_ok "$output"
+    else
+      msg_err "$output"
+      exit 1
+    fi
   fi
 
   printf "\nCreating empty commit with correct commit message.\n"
-  output="$(commit_message | git commit --allow-empty --file 2>&1 -)" &&
-    msg_ok "${output}" ||
-    (msg_err "${output}"; false)
+  if output="$(commit_message | git commit --allow-empty --file 2>&1 -)"; then
+    msg_ok "$output"
+  else
+    msg_err "$output"
+    exit 1
+  fi
 
   local ret=0
   if test -n "$try_apply" ; then
@@ -330,7 +340,8 @@ stage_patch() {
     * Do this only for _related_ patches (otherwise it increases the
       size of the pull request, making it harder to review)
 
-  When you are done, try "%s -s" to create the pull request.
+  When you are done, try "%s -s" to create the pull request,
+  or "%s -s --draft" to create a draft pull request.
 
   See the wiki for more information:
     * https://github.com/neovim/neovim/wiki/Merging-patches-from-upstream-vim
@@ -339,17 +350,21 @@ stage_patch() {
 }
 
 gh_pr() {
-  gh pr create --title "$1" --body "$2"
+  local pr_title
+  local pr_body
+  pr_title="$1"
+  pr_body="$2"
+  shift 2
+  gh pr create --title "${pr_title}" --body "${pr_body}" "$@"
 }
 
 git_hub_pr() {
   local pr_message
   pr_message="$(printf '%s\n\n%s\n' "$1" "$2")"
-  git hub pull new -m "${pr_message}"
+  shift 2
+  git hub pull new -m "${pr_message}" "$@"
 }
 
-# shellcheck disable=SC2015
-# ^ "Note that A && B || C is not if-then-else."
 submit_pr() {
   require_executable git
   local push_first
@@ -400,17 +415,23 @@ submit_pr() {
       fi
     fi
     echo "Pushing to '${push_remote}/${checked_out_branch}'."
-    output="$(git push "${push_remote}" "${checked_out_branch}" 2>&1)" &&
-      msg_ok "${output}" ||
-      (msg_err "${output}"; false)
+    if output="$(git push "$push_remote" "$checked_out_branch" 2>&1)"; then
+      msg_ok "$output"
+    else
+      msg_err "$output"
+      exit 1
+    fi
 
     echo
   fi
 
   echo "Creating pull request."
-  output="$(${submit_fn} "${pr_title}" "${pr_body}" 2>&1)" &&
-    msg_ok "${output}" ||
-    (msg_err "${output}"; false)
+  if output="$($submit_fn "$pr_title" "$pr_body" "$@" 2>&1)"; then
+    msg_ok "$output"
+  else
+    msg_err "$output"
+    exit 1
+  fi
 
   echo
   echo "Cleaning up files."
@@ -700,14 +721,14 @@ review_commit() {
   message_length="$(wc -l <<< "${expected_commit_message}")"
   local commit_message
   commit_message="$(tail -n +4 "${NVIM_SOURCE_DIR}/n${patch_file}" | head -n "${message_length}")"
-  if [[ "${commit_message#${git_patch_prefix}}" == "${expected_commit_message}" ]]; then
+  if [[ "${commit_message#"$git_patch_prefix"}" == "${expected_commit_message}" ]]; then
     msg_ok "Found expected commit message."
   else
     msg_err "Wrong commit message."
     echo "  Expected:"
     echo "${expected_commit_message}"
     echo "  Actual:"
-    echo "${commit_message#${git_patch_prefix}}"
+    echo "${commit_message#"$git_patch_prefix"}"
   fi
 
   get_vimpatch "${vim_version}"
@@ -796,7 +817,8 @@ while getopts "hlLmMVp:P:g:r:s" opt; do
       exit 0
       ;;
     s)
-      submit_pr
+      shift  # remove opt
+      submit_pr "$@"
       exit 0
       ;;
     V)
