@@ -4,23 +4,11 @@
 #include "nvim/lib/kvec.h"
 #include "nvim/types.h"
 
-typedef struct Decoration Decoration;
-
 typedef struct {
   char *text;
   int hl_id;
 } VirtTextChunk;
 
-
-typedef struct {
-  uint64_t ns_id;
-  uint64_t mark_id;
-  // TODO(bfredl): a lot of small allocations. Should probably use
-  // kvec_t(Decoration) as an arena. Alternatively, store ns_id/mark_id
-  // _inline_ in MarkTree and use the map only for decorations.
-  Decoration *decor;
-} ExtmarkItem;
-
 typedef struct undo_object ExtmarkUndoObject;
 typedef kvec_t(ExtmarkUndoObject) extmark_undo_vec_t;
 
