@@ -14,10 +14,10 @@
  */
 typedef struct foldinfo {
   linenr_T fi_lnum;             // line number where fold starts
-  int fi_level;                 /* level of the fold; when this is zero the
-                                   other fields are invalid */
-  int fi_low_level;             /* lowest fold level that starts in the same
-                                   line */
+  int fi_level;                 // level of the fold; when this is zero the
+                                // other fields are invalid
+  int fi_low_level;             // lowest fold level that starts in the same
+                                // line
   long fi_lines;
 } foldinfo_T;
 
