@@ -122,8 +122,6 @@
 //
 // Entries must be in the range 0x02-0x7f (see comment at K_SPECIAL).
 enum key_extra {
-  KE_NAME = 3,                 // name of this terminal entry
-
   KE_S_UP = 4,              // shift-up
   KE_S_DOWN = 5,             // shift-down
 
@@ -220,7 +218,7 @@ enum key_extra {
   KE_KINS = 79,             // keypad Insert key
   KE_KDEL = 80,              // keypad Delete key
 
-  KE_CSI = 81,              // CSI typed directly
+  // KE_CSI = 81,           // Nvim doesn't need escaping CSI
   KE_SNR = 82,              // <SNR>
   KE_PLUG = 83,             // <Plug>
   KE_CMDWIN = 84,            // open command-line window from Command-line Mode
@@ -245,6 +243,7 @@ enum key_extra {
   KE_MOUSEMOVE = 100,        // mouse moved with no button down
   // , KE_CANCEL = 101        // return from vgetc
   KE_EVENT = 102,           // event
+  KE_LUA = 103,  // lua special key
   KE_COMMAND = 104,  // <Cmd> special key
 };
 
@@ -326,6 +325,35 @@ enum key_extra {
 #define K_F35           TERMCAP2KEY('F', 'P')
 #define K_F36           TERMCAP2KEY('F', 'Q')
 #define K_F37           TERMCAP2KEY('F', 'R')
+#define K_F38           TERMCAP2KEY('F', 'S')
+#define K_F39           TERMCAP2KEY('F', 'T')
+#define K_F40           TERMCAP2KEY('F', 'U')
+
+#define K_F41           TERMCAP2KEY('F', 'V')
+#define K_F42           TERMCAP2KEY('F', 'W')
+#define K_F43           TERMCAP2KEY('F', 'X')
+#define K_F44           TERMCAP2KEY('F', 'Y')
+#define K_F45           TERMCAP2KEY('F', 'Z')
+#define K_F46           TERMCAP2KEY('F', 'a')
+#define K_F47           TERMCAP2KEY('F', 'b')
+#define K_F48           TERMCAP2KEY('F', 'c')
+#define K_F49           TERMCAP2KEY('F', 'd')
+#define K_F50           TERMCAP2KEY('F', 'e')
+
+#define K_F51           TERMCAP2KEY('F', 'f')
+#define K_F52           TERMCAP2KEY('F', 'g')
+#define K_F53           TERMCAP2KEY('F', 'h')
+#define K_F54           TERMCAP2KEY('F', 'i')
+#define K_F55           TERMCAP2KEY('F', 'j')
+#define K_F56           TERMCAP2KEY('F', 'k')
+#define K_F57           TERMCAP2KEY('F', 'l')
+#define K_F58           TERMCAP2KEY('F', 'm')
+#define K_F59           TERMCAP2KEY('F', 'n')
+#define K_F60           TERMCAP2KEY('F', 'o')
+
+#define K_F61           TERMCAP2KEY('F', 'p')
+#define K_F62           TERMCAP2KEY('F', 'q')
+#define K_F63           TERMCAP2KEY('F', 'r')
 
 // extra set of shifted function keys F1-F4, for vt100 compatible xterm
 #define K_S_XF1         TERMCAP2KEY(KS_EXTRA, KE_S_XF1)
@@ -434,7 +462,6 @@ enum key_extra {
 #define K_MOUSELEFT     TERMCAP2KEY(KS_EXTRA, KE_MOUSELEFT)
 #define K_MOUSERIGHT    TERMCAP2KEY(KS_EXTRA, KE_MOUSERIGHT)
 
-#define K_CSI           TERMCAP2KEY(KS_EXTRA, KE_CSI)
 #define K_SNR           TERMCAP2KEY(KS_EXTRA, KE_SNR)
 #define K_PLUG          TERMCAP2KEY(KS_EXTRA, KE_PLUG)
 #define K_CMDWIN        TERMCAP2KEY(KS_EXTRA, KE_CMDWIN)
@@ -443,6 +470,7 @@ enum key_extra {
 
 #define K_EVENT         TERMCAP2KEY(KS_EXTRA, KE_EVENT)
 #define K_COMMAND       TERMCAP2KEY(KS_EXTRA, KE_COMMAND)
+#define K_LUA           TERMCAP2KEY(KS_EXTRA, KE_LUA)
 
 // Bits for modifier mask
 // 0x01 cannot be used, because the modifier must be 0x02 or higher
