@@ -129,6 +129,10 @@ void tslua_init(lua_State *L)
   build_meta(L, TS_META_QUERY, query_meta);
   build_meta(L, TS_META_QUERYCURSOR, querycursor_meta);
   build_meta(L, TS_META_TREECURSOR, treecursor_meta);
+
+#ifdef NVIM_TS_HAS_SET_ALLOCATOR
+  ts_set_allocator(xmalloc, xcalloc, xrealloc, xfree);
+#endif
 }
 
 int tslua_has_language(lua_State *L)
@@ -1118,7 +1122,7 @@ static int node_rawquery(lua_State *L)
   // TODO(clason): API introduced after tree-sitter release 0.19.5
   // remove guard when minimum ts version is bumped to 0.19.6+
 #ifdef NVIM_TS_HAS_SET_MATCH_LIMIT
-  ts_query_cursor_set_match_limit(cursor, 32);
+  ts_query_cursor_set_match_limit(cursor, 64);
 #endif
   ts_query_cursor_exec(cursor, query, node);
 
