@@ -42,13 +42,13 @@
 #include "nvim/eval.h"
 #include "nvim/fileio.h"
 #include "nvim/func_attr.h"
+#include "nvim/getchar.h"
 #include "nvim/iconv.h"
 #include "nvim/mark.h"
 #include "nvim/mbyte.h"
 #include "nvim/memline.h"
 #include "nvim/memory.h"
 #include "nvim/message.h"
-#include "nvim/misc1.h"
 #include "nvim/option.h"
 #include "nvim/os/os.h"
 #include "nvim/path.h"
@@ -1067,7 +1067,7 @@ bool utf_printable(int c)
   static struct interval nonprint[] =
   {
     { 0x070f, 0x070f }, { 0x180b, 0x180e }, { 0x200b, 0x200f }, { 0x202a, 0x202e },
-    { 0x206a, 0x206f }, { 0xd800, 0xdfff }, { 0xfeff, 0xfeff }, { 0xfff9, 0xfffb },
+    { 0x2060, 0x206f }, { 0xd800, 0xdfff }, { 0xfeff, 0xfeff }, { 0xfff9, 0xfffb },
     { 0xfffe, 0xffff }
   };
 
@@ -1317,6 +1317,12 @@ bool mb_isupper(int a)
   return mb_tolower(a) != a;
 }
 
+bool mb_isalpha(int a)
+  FUNC_ATTR_WARN_UNUSED_RESULT
+{
+  return mb_islower(a) || mb_isupper(a);
+}
+
 static int utf_strnicmp(const char_u *s1, const char_u *s2, size_t n1, size_t n2)
 {
   int c1, c2, cdiff;
@@ -1609,7 +1615,8 @@ void show_utf8(void)
   msg((char *)IObuff);
 }
 
-/// Return offset from "p" to the first byte of the character it points into.
+/// Return offset from "p" to the start of a character, including composing characters.
+/// "base" must be the start of the string, which must be NUL terminated.
 /// If "p" points to the NUL at the end of the string return 0.
 /// Returns 0 when already at the first byte of a character.
 int utf_head_off(const char_u *base, const char_u *p)
@@ -1820,12 +1827,10 @@ void mb_copy_char(const char_u **const fp, char_u **const tp)
   *fp += l;
 }
 
-/*
- * Return the offset from "p" to the first byte of a character.  When "p" is
- * at the start of a character 0 is returned, otherwise the offset to the next
- * character.  Can start anywhere in a stream of bytes.
- */
-int mb_off_next(char_u *base, char_u *p)
+/// Return the offset from "p" to the first byte of a character.  When "p" is
+/// at the start of a character 0 is returned, otherwise the offset to the next
+/// character.  Can start anywhere in a stream of bytes.
+int mb_off_next(const char_u *base, const char_u *p)
 {
   int i;
   int j;
@@ -1850,11 +1855,10 @@ int mb_off_next(char_u *base, char_u *p)
   return i;
 }
 
-/*
- * Return the offset from "p" to the last byte of the character it points
- * into.  Can start anywhere in a stream of bytes.
- */
-int mb_tail_off(char_u *base, char_u *p)
+/// Return the offset from "p" to the last byte of the character it points
+/// into.  Can start anywhere in a stream of bytes.
+/// Composing characters are not included.
+int mb_tail_off(const char_u *base, const char_u *p)
 {
   int i;
   int j;
@@ -1882,12 +1886,13 @@ int mb_tail_off(char_u *base, char_u *p)
 
 /// Return the offset from "p" to the first byte of the character it points
 /// into. Can start anywhere in a stream of bytes.
+/// Unlike utf_head_off() this doesn't include composing characters and returns a negative value.
 ///
 /// @param[in] base  Pointer to start of string
 /// @param[in] p     Pointer to byte for which to return the offset to the previous codepoint
 //
 /// @return 0 if invalid sequence, else offset to previous codepoint
-int mb_head_off(char_u *base, char_u *p)
+int mb_head_off(const char_u *base, const char_u *p)
 {
   int i;
   int j;
@@ -2037,13 +2042,11 @@ char_u *mb_prevptr(char_u *line, char_u *p)
   return p;
 }
 
-/*
- * Return the character length of "str".  Each multi-byte character (with
- * following composing characters) counts as one.
- */
-int mb_charlen(char_u *str)
+/// Return the character length of "str".  Each multi-byte character (with
+/// following composing characters) counts as one.
+int mb_charlen(const char_u *str)
 {
-  char_u *p = str;
+  const char_u *p = str;
   int count;
 
   if (p == NULL) {
@@ -2057,12 +2060,10 @@ int mb_charlen(char_u *str)
   return count;
 }
 
-/*
- * Like mb_charlen() but for a string with specified length.
- */
-int mb_charlen_len(char_u *str, int len)
+/// Like mb_charlen() but for a string with specified length.
+int mb_charlen_len(const char_u *str, int len)
 {
-  char_u *p = str;
+  const char_u *p = str;
   int count;
 
   for (count = 0; *p != NUL && p < str + len; count++) {
@@ -2089,8 +2090,7 @@ const char *mb_unescape(const char **const pp)
   size_t buf_idx = 0;
   uint8_t *str = (uint8_t *)(*pp);
 
-  // Must translate K_SPECIAL KS_SPECIAL KE_FILLER to K_SPECIAL and CSI
-  // KS_EXTRA KE_CSI to CSI.
+  // Must translate K_SPECIAL KS_SPECIAL KE_FILLER to K_SPECIAL.
   // Maximum length of a utf-8 character is 4 bytes.
   for (size_t str_idx = 0; str[str_idx] != NUL && buf_idx < 4; str_idx++) {
     if (str[str_idx] == K_SPECIAL
@@ -2098,11 +2098,6 @@ const char *mb_unescape(const char **const pp)
         && str[str_idx + 2] == KE_FILLER) {
       buf[buf_idx++] = (char)K_SPECIAL;
       str_idx += 2;
-    } else if ((str[str_idx] == K_SPECIAL)
-               && str[str_idx + 1] == KS_EXTRA
-               && str[str_idx + 2] == KE_CSI) {
-      buf[buf_idx++] = (char)CSI;
-      str_idx += 2;
     } else if (str[str_idx] == K_SPECIAL) {
       break;  // A special key can't be a multibyte char.
     } else {
@@ -2207,11 +2202,9 @@ char_u *enc_canonize(char_u *enc) FUNC_ATTR_NONNULL_RET
   return r;
 }
 
-/*
- * Search for an encoding alias of "name".
- * Returns -1 when not found.
- */
-static int enc_alias_search(char_u *name)
+/// Search for an encoding alias of "name".
+/// Returns -1 when not found.
+static int enc_alias_search(const char_u *name)
 {
   int i;
 
