@@ -274,6 +274,14 @@ return {
       varname='p_cmp',
       defaults={if_true="internal,keepascii"}
     },
+    {
+      full_name='cdhome', abbreviation='cdh',
+      short_desc=N_(":cd without argument goes to the home directory"),
+      type='bool', scope={'global'},
+      secure=true,
+      varname='p_cdh',
+      defaults={if_true=false}
+    },
     {
       full_name='cdpath', abbreviation='cd',
       short_desc=N_("list of directories searched with \":cd\""),
@@ -342,6 +350,15 @@ return {
       varname='p_cinw',
       defaults={if_true="if,else,while,do,for,switch"}
     },
+    {
+      full_name='cinscopedecls', abbreviation='cinsd',
+      short_desc=N_("words that are recognized by 'cino-g'"),
+      type='string', list='onecomma', scope={'buffer'},
+      deny_duplicates=true,
+      alloced=true,
+      varname='p_cinsd',
+      defaults={if_true="public,protected,private"}
+    },
     {
       full_name='clipboard', abbreviation='cb',
       short_desc=N_("use the clipboard as the unnamed register"),
@@ -403,7 +420,7 @@ return {
     },
     {
       full_name='compatible', abbreviation='cp',
-	  short_desc=N_("No description"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
       redraw={'all_windows'},
       varname='p_force_off',
@@ -657,14 +674,14 @@ return {
     },
     {
       full_name='edcompatible', abbreviation='ed',
-	  short_desc=N_("No description"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
       varname='p_force_off',
       defaults={if_true=false}
     },
     {
       full_name='emoji', abbreviation='emo',
-	  short_desc=N_("No description"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
       redraw={'all_windows', 'ui_option'},
       varname='p_emoji',
@@ -1176,7 +1193,7 @@ return {
     },
     {
       full_name='inccommand', abbreviation='icm',
-	  short_desc=N_("Live preview of substitution"),
+      short_desc=N_("Live preview of substitution"),
       type='string', scope={'global'},
       redraw={'all_windows'},
       varname='p_icm',
@@ -1838,7 +1855,7 @@ return {
       type='number', scope={'global'},
       secure=true,
       varname='p_pyx',
-      defaults={if_true=0}
+      defaults={if_true=3}
     },
     {
       full_name='quickfixtextfunc', abbreviation='qftf',
@@ -2491,7 +2508,7 @@ return {
     },
     {
       full_name='termencoding', abbreviation='tenc',
-      short_desc=N_("Terminal encodig"),
+      short_desc=N_("Terminal encoding"),
       type='string', scope={'global'},
       defaults={if_true=""}
     },
@@ -2614,7 +2631,7 @@ return {
     },
     {
       full_name='ttyfast', abbreviation='tf',
-	  short_desc=N_("No description"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
       no_mkrc=true,
       varname='p_force_on',
@@ -2728,7 +2745,7 @@ return {
     {
       full_name='virtualedit', abbreviation='ve',
       short_desc=N_("when to use virtual editing"),
-      type='string', list='onecomma', scope={'global'},
+      type='string', list='onecomma', scope={'global', 'window'},
       deny_duplicates=true,
       redraw={'curswant'},
       varname='p_ve',
