@@ -31,7 +31,6 @@
 #include "nvim/memline.h"
 #include "nvim/memory.h"
 #include "nvim/message.h"
-#include "nvim/misc1.h"
 #include "nvim/move.h"
 #include "nvim/ops.h"
 #include "nvim/option.h"
@@ -355,7 +354,7 @@ char *strcase_save(const char *const orig, bool upper)
     int l = utf_ptr2len((const char_u *)p);
     if (c == 0) {
       // overlong sequence, use only the first byte
-      c = *p;
+      c = (char_u)(*p);
       l = 1;
     }
     int uc = upper ? mb_toupper(c) : mb_tolower(c);
@@ -394,6 +393,18 @@ void del_trailing_spaces(char_u *ptr)
   }
 }
 
+#if !defined(HAVE_STRNLEN)
+size_t xstrnlen(const char *s, size_t n)
+  FUNC_ATTR_NONNULL_ALL FUNC_ATTR_PURE
+{
+  const char *end = memchr(s, '\0', n);
+  if (end == NULL) {
+    return n;
+  }
+  return end - s;
+}
+#endif
+
 #if (!defined(HAVE_STRCASECMP) && !defined(HAVE_STRICMP))
 /*
  * Compare two strings, ignoring case, using current locale.
@@ -990,22 +1001,20 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
                                  - str_arg);
           }
           if (fmt_spec == 'S') {
-            if (min_field_width != 0) {
-              min_field_width += (strlen(str_arg)
-                                  - mb_string2cells((char_u *)str_arg));
-            }
-            if (precision) {
-              char_u *p1;
-              size_t i = 0;
+            char_u *p1;
+            size_t i;
 
-              for (p1 = (char_u *)str_arg; *p1;
-                   p1 += utfc_ptr2len(p1)) {
-                i += (size_t)utf_ptr2cells(p1);
-                if (i > precision) {
-                  break;
-                }
+            for (i = 0, p1 = (char_u *)str_arg; *p1; p1 += utfc_ptr2len(p1)) {
+              size_t cell = (size_t)utf_ptr2cells(p1);
+              if (precision_specified && i + cell > precision) {
+                break;
               }
-              str_arg_l = (size_t)(p1 - (char_u *)str_arg);
+              i += cell;
+            }
+
+            str_arg_l = (size_t)(p1 - (char_u *)str_arg);
+            if (min_field_width != 0) {
+              min_field_width += str_arg_l - i;
             }
           }
           break;
