@@ -53,6 +53,7 @@ describe('startup', function()
     ]])
   end)
   it('in a TTY: has("ttyin")==1 has("ttyout")==1', function()
+    if helpers.pending_win32(pending) then return end
     local screen = Screen.new(25, 4)
     screen:attach()
     if iswin() then
@@ -104,6 +105,7 @@ describe('startup', function()
     end)
   end)
   it('input from pipe (implicit) #7679', function()
+    if helpers.pending_win32(pending) then return end
     local screen = Screen.new(25, 4)
     screen:attach()
     if iswin() then
@@ -224,6 +226,23 @@ describe('startup', function()
     end
   end)
 
+  it('-e sets ex mode', function()
+    local screen = Screen.new(25, 3)
+    clear('-e')
+    screen:attach()
+    -- Verify we set the proper mode both before and after :vi.
+    feed("put =mode(1)<CR>vi<CR>:put =mode(1)<CR>")
+    screen:expect([[
+      cv                       |
+      ^n                        |
+      :put =mode(1)            |
+    ]])
+
+    eq('cv\n',
+       funcs.system({nvim_prog, '-n', '-es' },
+                    { 'put =mode(1)', 'print', '' }))
+  end)
+
   it('fails on --embed with -es/-Es', function()
     matches('nvim[.exe]*: %-%-embed conflicts with %-es/%-Es',
       funcs.system({nvim_prog, '--embed', '-es' }))
@@ -242,6 +261,7 @@ describe('startup', function()
   end)
 
   it('ENTER dismisses early message #7967', function()
+    if helpers.pending_win32(pending) then return end
     local screen
     screen = Screen.new(60, 6)
     screen:attach()
@@ -338,7 +358,7 @@ describe('startup', function()
   end)
 
   it("handles :packadd during startup", function()
-    -- control group: opt/bonus is not availabe by default
+    -- control group: opt/bonus is not available by default
     pack_clear [[
       try
         let g:x = bonus#secret()
@@ -474,6 +494,7 @@ describe('sysinit', function()
   end)
 
   it('fixed hang issue with -D (#12647)', function()
+    if helpers.pending_win32(pending) then return end
     local screen
     screen = Screen.new(60, 6)
     screen:attach()
@@ -651,11 +672,7 @@ describe('runtime:', function()
     mkdir_p(ftdetect_folder)
     write_file(ftdetect_file , [[vim.g.lua_ftdetect = 1]])
 
-    -- TODO(shadmansaleh): Figure out why this test fails without
-    --                     setting VIMRUNTIME
-    clear{ args_rm={'-u'}, env={XDG_CONFIG_HOME=xconfig,
-                                XDG_DATA_HOME=xdata,
-                                VIMRUNTIME='runtime/'}}
+    clear{ args_rm={'-u'}, env=xenv }
 
     eq(1, eval('g:lua_ftdetect'))
     rmdir(ftdetect_folder)
