@@ -5,7 +5,7 @@ local eq = helpers.eq
 local eval = helpers.eval
 local feed = helpers.feed
 local nvim = helpers.nvim
-local nvim_dir = helpers.nvim_dir
+local testprg = helpers.testprg
 local retry = helpers.retry
 
 describe(':ls', function()
@@ -14,7 +14,7 @@ describe(':ls', function()
   end)
 
   it('R, F for :terminal buffers', function()
-    nvim('set_option', 'shell', string.format('"%s" INTERACT', nvim_dir..'/shell-test'))
+    nvim('set_option', 'shell', string.format('"%s" INTERACT', testprg('shell-test')))
 
     command('edit foo')
     command('set hidden')
