@@ -119,8 +119,10 @@ function tests.check_workspace_configuration()
       notify('workspace/configuration', { items = {
               { section = "testSetting1" };
               { section = "testSetting2" };
+              { section = "test.Setting3" };
+              { section = "test.Setting4" };
           } })
-      expect_notification('workspace/configuration', { true; vim.NIL})
+      expect_notification('workspace/configuration', { true; false; 'nested'; vim.NIL})
       notify('shutdown')
     end;
   }
@@ -671,6 +673,34 @@ function tests.code_action_with_resolve()
   }
 end
 
+function tests.code_action_server_side_command()
+  skeleton({
+    on_init = function()
+      return {
+        capabilities = {
+          codeActionProvider = {
+            resolveProvider = false,
+          },
+        },
+      }
+    end,
+    body = function()
+      notify('start')
+      local cmd = {
+        title = 'Command 1',
+        command = 'dummy1',
+      }
+      expect_request('textDocument/codeAction', function()
+        return nil, { cmd }
+      end)
+      expect_request('workspace/executeCommand', function()
+        return nil, cmd
+      end)
+      notify('shutdown')
+    end,
+  })
+end
+
 function tests.clientside_commands()
   skeleton {
     on_init = function()
@@ -685,6 +715,49 @@ function tests.clientside_commands()
   }
 end
 
+function tests.codelens_refresh_lock()
+  skeleton {
+    on_init = function()
+      return {
+        capabilities = {
+          codeLensProvider = { resolveProvider = true; };
+        }
+      }
+    end;
+    body = function()
+      notify('start')
+      expect_request("textDocument/codeLens", function ()
+        return {code = -32002, message = "ServerNotInitialized"}, nil
+      end)
+      expect_request("textDocument/codeLens", function ()
+        local lenses = {
+          {
+            range = {
+              start = { line = 0, character = 0, },
+              ['end'] = { line = 0, character = 3 }
+            },
+            command = { title = 'Lens1', command = 'Dummy' }
+          },
+        }
+        return nil, lenses
+      end)
+      expect_request("textDocument/codeLens", function ()
+        local lenses = {
+          {
+            range = {
+              start = { line = 0, character = 0, },
+              ['end'] = { line = 0, character = 3 }
+            },
+            command = { title = 'Lens2', command = 'Dummy' }
+          },
+        }
+        return nil, lenses
+      end)
+      notify('shutdown')
+    end;
+  }
+end
+
 -- Tests will be indexed by TEST_NAME
 
 local kill_timer = vim.loop.new_timer()
