@@ -46,7 +46,7 @@ describe('eval', function()
     command('AR "')
     command([[let @" = "abc\n"]])
     source('AR "')
-    command([[let @" = "abc\<C-m>"]])
+    command([[let @" = "abc\r"]])
     command('AR "')
     command([[let @= = '"abc"']])
     command('AR =')
@@ -666,7 +666,7 @@ describe('eval', function()
     source([[
       " Vim script used in test_eval.in.  Needed for script-local function.
 
-      func! s:Testje()
+      func s:Testje()
         return "foo"
       endfunc
 
