@@ -252,9 +252,8 @@ describe('lua buffer event callbacks: on_lines', function()
     eq(2, meths.win_get_cursor(0)[1])
   end)
 
-  it('does not SEGFAULT when calling win_findbuf in on_detach', function()
-
-    exec_lua[[
+  it('does not SEGFAULT when accessing window buffer info in on_detach #14998', function()
+    local code = [[
       local buf = vim.api.nvim_create_buf(false, false)
 
       vim.cmd"split"
@@ -262,13 +261,19 @@ describe('lua buffer event callbacks: on_lines', function()
 
       vim.api.nvim_buf_attach(buf, false, {
         on_detach = function(_, buf)
+          vim.fn.tabpagebuflist()
           vim.fn.win_findbuf(buf)
         end
       })
     ]]
 
+    exec_lua(code)
     command("q!")
     helpers.assert_alive()
+
+    exec_lua(code)
+    command("bd!")
+    helpers.assert_alive()
   end)
 
   it('#12718 lnume', function()
@@ -1104,6 +1109,15 @@ describe('lua: nvim_buf_attach on_bytes', function()
       check_events { }
     end)
 
+    it("works with accepting spell suggestions", function()
+      local check_events = setup_eventcheck(verify, {"hallo"})
+
+      feed("gg0z=4<cr><cr>") -- accepts 'Hello'
+      check_events {
+        { "test1", "bytes", 1, 3, 0, 0, 0, 0, 2, 2, 0, 2, 2 };
+      }
+    end)
+
     local function test_lockmarks(mode)
       local description = (mode ~= "") and mode or "(baseline)"
       it("test_lockmarks " .. description .. " %delete _", function()
