@@ -208,10 +208,10 @@ describe('vim.diagnostic', function()
     eq(all_highlights, exec_lua [[
       local ns_1_diags = {
         make_error("Error 1", 1, 1, 1, 5),
-        make_warning("Warning on Server 1", 2, 1, 2, 5),
+        make_warning("Warning on Server 1", 2, 1, 2, 3),
       }
       local ns_2_diags = {
-        make_warning("Warning 1", 2, 1, 2, 5),
+        make_warning("Warning 1", 2, 1, 2, 3),
       }
 
       vim.diagnostic.set(diagnostic_ns, diagnostic_bufnr, ns_1_diags)
@@ -255,10 +255,10 @@ describe('vim.diagnostic', function()
     eq({0, 2}, exec_lua [[
       local ns_1_diags = {
         make_error("Error 1", 1, 1, 1, 5),
-        make_warning("Warning on Server 1", 2, 1, 2, 5),
+        make_warning("Warning on Server 1", 2, 1, 2, 3),
       }
       local ns_2_diags = {
-        make_warning("Warning 1", 2, 1, 2, 5),
+        make_warning("Warning 1", 2, 1, 2, 3),
       }
 
       vim.diagnostic.set(diagnostic_ns, diagnostic_bufnr, ns_1_diags)
@@ -599,10 +599,10 @@ describe('vim.diagnostic', function()
       eq(all_highlights, exec_lua [[
         local ns_1_diags = {
           make_error("Error 1", 1, 1, 1, 5),
-          make_warning("Warning on Server 1", 2, 1, 2, 5),
+          make_warning("Warning on Server 1", 2, 1, 2, 3),
         }
         local ns_2_diags = {
-          make_warning("Warning 1", 2, 1, 2, 5),
+          make_warning("Warning 1", 2, 1, 2, 3),
         }
 
         vim.diagnostic.set(diagnostic_ns, diagnostic_bufnr, ns_1_diags)
@@ -787,7 +787,7 @@ describe('vim.diagnostic', function()
       eq(2, exec_lua [[
         vim.diagnostic.set(diagnostic_ns, diagnostic_bufnr, {
           make_error("Error 1", 1, 1, 1, 5),
-          make_warning("Warning on Server 1", 1, 1, 2, 5),
+          make_warning("Warning on Server 1", 1, 1, 2, 3),
         })
 
         return #vim.diagnostic.get(diagnostic_bufnr)
@@ -798,9 +798,9 @@ describe('vim.diagnostic', function()
       eq({2, 3, 2}, exec_lua [[
         vim.diagnostic.set(diagnostic_ns, diagnostic_bufnr, {
           make_error("Error 1", 1, 1, 1, 5),
-          make_warning("Warning on Server 1", 1, 1, 2, 5),
-          make_info("Ignored information", 1, 1, 2, 5),
-          make_hint("Here's a hint", 1, 1, 2, 5),
+          make_warning("Warning on Server 1", 1, 1, 2, 3),
+          make_info("Ignored information", 1, 1, 2, 3),
+          make_hint("Here's a hint", 1, 1, 2, 3),
         })
 
         return {
@@ -820,8 +820,8 @@ describe('vim.diagnostic', function()
       eq(1, exec_lua [[
         vim.diagnostic.set(diagnostic_ns, diagnostic_bufnr, {
           make_error("Error 1", 1, 1, 1, 5),
-          make_warning("Warning on Server 1", 1, 1, 2, 5),
-          make_info("Ignored information", 1, 1, 2, 5),
+          make_warning("Warning on Server 1", 1, 1, 2, 3),
+          make_info("Ignored information", 1, 1, 2, 3),
           make_error("Error On Other Line", 2, 1, 1, 5),
         })
 
@@ -1119,6 +1119,11 @@ describe('vim.diagnostic', function()
   end)
 
   describe('set()', function()
+    it('validates its arguments', function()
+      matches("expected a list of diagnostics",
+        pcall_err(exec_lua, [[vim.diagnostic.set(1, 0, {lnum = 1, col = 2})]]))
+    end)
+
     it('can perform updates after insert_leave', function()
       exec_lua [[vim.api.nvim_set_current_buf(diagnostic_bufnr)]]
       nvim("input", "o")
@@ -1934,24 +1939,31 @@ describe('vim.diagnostic', function()
     end)
 
     it('triggers the autocommand when diagnostics are set', function()
-      eq(1, exec_lua [[
+      eq(true, exec_lua [[
+        -- Set a different buffer as current to test that <abuf> is being set properly in
+        -- DiagnosticChanged callbacks
+        local tmp = vim.api.nvim_create_buf(false, true)
+        vim.api.nvim_set_current_buf(tmp)
+
         vim.g.diagnostic_autocmd_triggered = 0
-        vim.cmd('autocmd DiagnosticChanged * let g:diagnostic_autocmd_triggered = 1')
+        vim.cmd('autocmd DiagnosticChanged * let g:diagnostic_autocmd_triggered = +expand("<abuf>")')
         vim.api.nvim_buf_set_name(diagnostic_bufnr, "test | test")
         vim.diagnostic.set(diagnostic_ns, diagnostic_bufnr, {
           make_error('Diagnostic', 0, 0, 0, 0)
         })
-        return vim.g.diagnostic_autocmd_triggered
+        return vim.g.diagnostic_autocmd_triggered == diagnostic_bufnr
       ]])
       end)
 
     it('triggers the autocommand when diagnostics are cleared', function()
-      eq(1, exec_lua [[
+      eq(true, exec_lua [[
+        local tmp = vim.api.nvim_create_buf(false, true)
+        vim.api.nvim_set_current_buf(tmp)
         vim.g.diagnostic_autocmd_triggered = 0
-        vim.cmd('autocmd DiagnosticChanged * let g:diagnostic_autocmd_triggered = 1')
+        vim.cmd('autocmd DiagnosticChanged * let g:diagnostic_autocmd_triggered = +expand("<abuf>")')
         vim.api.nvim_buf_set_name(diagnostic_bufnr, "test | test")
         vim.diagnostic.reset(diagnostic_ns, diagnostic_bufnr)
-        return vim.g.diagnostic_autocmd_triggered
+        return vim.g.diagnostic_autocmd_triggered == diagnostic_bufnr
       ]])
       end)
   end)
