@@ -14,7 +14,7 @@ local feed_command = helpers.feed_command
 local feed_data = thelpers.feed_data
 local clear = helpers.clear
 local command = helpers.command
-local nvim_dir = helpers.nvim_dir
+local testprg = helpers.testprg
 local retry = helpers.retry
 local nvim_prog = helpers.nvim_prog
 local nvim_set = helpers.nvim_set
@@ -214,7 +214,7 @@ describe('TUI', function()
     ]])
   end)
 
-  it('interprets ESC+key as ALT chord', function()
+  it('interprets ESC+key as ALT chord in i_CTRL-V', function()
     -- Vim represents ALT/META by setting the "high bit" of the modified key:
     -- ALT+j inserts "ê". Nvim does not (#3982).
     feed_data('i\022\027j')
@@ -229,6 +229,38 @@ describe('TUI', function()
     ]])
   end)
 
+  it('interprets <Esc>[27u as <Esc>', function()
+    feed_command('nnoremap <M-;> <Nop>')
+    feed_command('nnoremap <Esc> AESC<Esc>')
+    feed_command('nnoremap ; Asemicolon<Esc>')
+    feed_data('\027[27u;')
+    screen:expect([[
+      ESCsemicolo{1:n}                                      |
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {5:[No Name] [+]                                     }|
+                                                        |
+      {3:-- TERMINAL --}                                    |
+    ]])
+    -- <Esc>; should be recognized as <M-;> when <M-;> is mapped
+    feed_data('\027;')
+    screen:expect_unchanged()
+  end)
+
+  it('interprets <Esc><Nul> as <M-C-Space> #17198', function()
+    feed_data('i\022\027\000')
+    screen:expect([[
+      <M-C-Space>{1: }                                      |
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {5:[No Name] [+]                                     }|
+      {3:-- INSERT --}                                      |
+      {3:-- TERMINAL --}                                    |
+    ]])
+  end)
+
   it('accepts ASCII control sequences', function()
     feed_data('i')
     feed_data('\022\007') -- ctrl+g
@@ -271,7 +303,7 @@ describe('TUI', function()
       {3:-- TERMINAL --}                                    |
     ]])
     feed_data('\027[201~')  -- End paste.
-    feed_data('\027\000')   -- ESC: go to Normal mode.
+    feed_data('\027[27u')   -- ESC: go to Normal mode.
     wait_for_mode('n')
     screen:expect([[
       "pasted from termina{1:l}"                            |
@@ -323,7 +355,7 @@ describe('TUI', function()
     feed_data('just paste it™')
     feed_data('\027[201~')
     screen:expect{grid=[[
-      thisjust paste it™{1:3} is here                       |
+      thisjust paste it{1:™}3 is here                       |
                                                         |
       {4:~                                                 }|
       {4:~                                                 }|
@@ -353,7 +385,7 @@ describe('TUI', function()
         return
     end
     feed_data(':set statusline=^^^^^^^\n')
-    feed_data(':terminal '..nvim_dir..'/tty-test\n')
+    feed_data(':terminal '..testprg('tty-test')..'\n')
     feed_data('i')
     screen:expect{grid=[[
       tty ready                                         |
@@ -379,7 +411,7 @@ describe('TUI', function()
   end)
 
   it('paste: normal-mode (+CRLF #10872)', function()
-    feed_data(':set ruler')
+    feed_data(':set ruler | echo')
     wait_for_mode('c')
     feed_data('\n')
     wait_for_mode('n')
@@ -423,13 +455,13 @@ describe('TUI', function()
     expect_child_buf_lines(expected_crlf)
     feed_data('u')
     expect_child_buf_lines({''})
+    feed_data(':echo')
+    wait_for_mode('c')
+    feed_data('\n')
+    wait_for_mode('n')
     -- CRLF input
     feed_data('\027[200~'..table.concat(expected_lf,'\r\n')..'\027[201~')
-    screen:expect{
-      grid=expected_grid1:gsub(
-        ':set ruler *',
-        '3 fewer lines; before #1  0 seconds ago           '),
-      attr_ids=expected_attr}
+    screen:expect{grid=expected_grid1, attr_ids=expected_attr}
     expect_child_buf_lines(expected_crlf)
   end)
 
@@ -453,7 +485,7 @@ describe('TUI', function()
       {3:-- TERMINAL --}                                    |
     ]]}
     -- Dot-repeat/redo.
-    feed_data('\027\000')
+    feed_data('\027[27u')
     wait_for_mode('n')
     feed_data('.')
     screen:expect{grid=[[
@@ -499,7 +531,7 @@ describe('TUI', function()
       vim.paste = function(lines, phase) error("fake fail") end
     ]], {})
     -- Prepare something for dot-repeat/redo.
-    feed_data('ifoo\n\027\000')
+    feed_data('ifoo\n\027[27u')
     wait_for_mode('n')
     screen:expect{grid=[[
       foo                                               |
@@ -541,7 +573,7 @@ describe('TUI', function()
       {3:-- TERMINAL --}                                    |
     ]]}
     -- Editor should still work after failed/drained paste.
-    feed_data('ityped input...\027\000')
+    feed_data('ityped input...\027[27u')
     screen:expect{grid=[[
       foo                                               |
       foo                                               |
@@ -575,23 +607,23 @@ describe('TUI', function()
       vim.paste = function(lines, phase) return false end
     ]], {})
     feed_data('\027[200~line A\nline B\n\027[201~')
-    feed_data('ifoo\n\027\000')
+    feed_data('ifoo\n\027[27u')
     expect_child_buf_lines({'foo',''})
   end)
 
   it("paste: 'nomodifiable' buffer", function()
     child_session:request('nvim_command', 'set nomodifiable')
     child_session:request('nvim_exec_lua', [[
-      -- Stack traces for this test are non-deterministic, so disable them
-      _G.debug.traceback = function(msg) return msg end
+      -- Truncate the error message to hide the line number
+      _G.debug.traceback = function(msg) return msg:sub(-49) end
     ]], {})
     feed_data('\027[200~fail 1\nfail 2\n\027[201~')
     screen:expect{grid=[[
                                                         |
       {4:~                                                 }|
       {5:                                                  }|
-      {8:paste: Error executing lua: vim.lua:243: Vim:E21: }|
-      {8:Cannot make changes, 'modifiable' is off}          |
+      {8:paste: Error executing lua: Vim:E21: Cannot make c}|
+      {8:hanges, 'modifiable' is off}                       |
       {10:Press ENTER or type command to continue}{1: }          |
       {3:-- TERMINAL --}                                    |
     ]]}
@@ -669,7 +701,7 @@ describe('TUI', function()
       {3:-- INSERT --}                                      |
       {3:-- TERMINAL --}                                    |
     ]])
-    feed_data('\027\000')  -- ESC: go to Normal mode.
+    feed_data('\027[27u')  -- ESC: go to Normal mode.
     wait_for_mode('n')
     -- Dot-repeat/redo.
     feed_data('.')
@@ -677,8 +709,8 @@ describe('TUI', function()
       item 2997                                         |
       item 2998                                         |
       item 2999                                         |
-      item 3000 en{1:d}                                     |
-      {5:[No Name] [+]                   3000,13        Bot}|
+      item 3000 en{1:d}d                                    |
+      {5:[No Name] [+]                   5999,13        Bot}|
                                                         |
       {3:-- TERMINAL --}                                    |
     ]])
@@ -765,6 +797,44 @@ describe('TUI', function()
     ]])
   end)
 
+  it('paste: streamed paste with isolated "stop paste" code', function()
+    child_session:request('nvim_exec_lua', [[
+      _G.paste_phases = {}
+      vim.paste = (function(overridden)
+        return function(lines, phase)
+          table.insert(_G.paste_phases, phase)
+          overridden(lines, phase)
+        end
+      end)(vim.paste)
+    ]], {})
+    feed_data('i')
+    feed_data('\027[200~pasted')  -- phase 1
+    screen:expect([[
+      pasted{1: }                                           |
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {5:[No Name] [+]                                     }|
+      {3:-- INSERT --}                                      |
+      {3:-- TERMINAL --}                                    |
+    ]])
+    feed_data(' from terminal')  -- phase 2
+    screen:expect([[
+      pasted from terminal{1: }                             |
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {5:[No Name] [+]                                     }|
+      {3:-- INSERT --}                                      |
+      {3:-- TERMINAL --}                                    |
+    ]])
+    -- Send isolated "stop paste" sequence.
+    feed_data('\027[201~')  -- phase 3
+    screen:expect_unchanged()
+    local _, rv = child_session:request('nvim_exec_lua', [[return _G.paste_phases]], {})
+    eq({1, 2, 3}, rv)
+  end)
+
   it('allows termguicolors to be set at runtime', function()
     screen:set_option('rgb', true)
     screen:set_default_attr_ids({
@@ -833,7 +903,7 @@ describe('TUI', function()
 
     feed_data(':set statusline=^^^^^^^\n')
     feed_data(':set termguicolors\n')
-    feed_data(':terminal '..nvim_dir..'/tty-test\n')
+    feed_data(':terminal '..testprg('tty-test')..'\n')
     -- Depending on platform the above might or might not fit in the cmdline
     -- so clear it for consistent behavior.
     feed_data(':\027')
@@ -1060,7 +1130,7 @@ describe('TUI FocusGained/FocusLost', function()
   end)
 
   it('in terminal-mode', function()
-    feed_data(':set shell='..nvim_dir..'/shell-test\n')
+    feed_data(':set shell='..testprg('shell-test')..'\n')
     feed_data(':set noshowmode laststatus=0\n')
 
     feed_data(':terminal\n')
