@@ -33,7 +33,7 @@ before_each(function()
     let g:NUM_LVLS = 4
     function Redraw()
       mode
-      return ''
+      return "\<Ignore>"
     endfunction
     let g:id = ''
     cnoremap <expr> {REDRAW} Redraw()
@@ -42,7 +42,7 @@ before_each(function()
       let Cb = g:Nvim_color_input{g:id}
       let out = input({'prompt': ':', 'highlight': Cb})
       let g:out{id} = out
-      return (a:do_return ? out : '')
+      return (a:do_return ? out : "\<Ignore>")
     endfunction
     nnoremap <expr> {PROMPT} DoPrompt(0)
     cnoremap <expr> {PROMPT} DoPrompt(1)
@@ -410,7 +410,7 @@ describe('Command-line coloring', function()
   end)
   it('stops executing callback after a number of errors', function()
     set_color_cb('SplittedMultibyteStart')
-    start_prompt('let x = "«»«»«»«»«»"\n')
+    start_prompt('let x = "«»«»«»«»«»"')
     screen:expect([[
       {EOB:~                                       }|
       {EOB:~                                       }|
@@ -419,7 +419,7 @@ describe('Command-line coloring', function()
       :let x = "                              |
       {ERR:E5405: Chunk 0 start 10 splits multibyte}|
       {ERR: character}                              |
-      ^:let x = "«»«»«»«»«»"                   |
+      :let x = "«»«»«»«»«»"^                   |
     ]])
     feed('\n')
     screen:expect([[
@@ -432,6 +432,7 @@ describe('Command-line coloring', function()
       {EOB:~                                       }|
                                               |
     ]])
+    feed('\n')
     eq('let x = "«»«»«»«»«»"', meths.get_var('out'))
     local msg = '\nE5405: Chunk 0 start 10 splits multibyte character'
     eq(msg:rep(1), funcs.execute('messages'))
@@ -474,14 +475,14 @@ describe('Command-line coloring', function()
     ]])
     feed('\n')
     screen:expect([[
-                                              |
+      ^                                        |
       {EOB:~                                       }|
       {EOB:~                                       }|
       {EOB:~                                       }|
       {EOB:~                                       }|
       {EOB:~                                       }|
       {EOB:~                                       }|
-      ^:echo 42                                |
+      :echo 42                                |
     ]])
     feed('\n')
     eq('echo 42', meths.get_var('out'))
