@@ -157,6 +157,99 @@ describe('Signs', function()
       ]])
     end)
 
+    it('higlights the cursorline sign with culhl', function()
+      feed('ia<cr>b<cr>c<esc>')
+      command('sign define piet text=>> texthl=Search culhl=ErrorMsg')
+      command('sign place 1 line=1 name=piet buffer=1')
+      command('sign place 2 line=2 name=piet buffer=1')
+      command('sign place 3 line=3 name=piet buffer=1')
+      command('set cursorline')
+      screen:expect([[
+        {1:>>}a                                                  |
+        {1:>>}b                                                  |
+        {8:>>}{3:^c                                                  }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+                                                             |
+      ]])
+      feed('k')
+      screen:expect([[
+        {1:>>}a                                                  |
+        {8:>>}{3:^b                                                  }|
+        {1:>>}c                                                  |
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+                                                             |
+      ]])
+      command('set nocursorline')
+      screen:expect([[
+        {1:>>}a                                                  |
+        {1:>>}^b                                                  |
+        {1:>>}c                                                  |
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+                                                             |
+      ]])
+      command('set cursorline cursorlineopt=line')
+      screen:expect([[
+        {1:>>}a                                                  |
+        {1:>>}{3:^b                                                  }|
+        {1:>>}c                                                  |
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+                                                             |
+      ]])
+      command('set cursorlineopt=number')
+      screen:expect([[
+        {1:>>}a                                                  |
+        {8:>>}^b                                                  |
+        {1:>>}c                                                  |
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+                                                             |
+      ]])
+    end)
+
     it('multiple signs #9295', function()
       feed('ia<cr>b<cr>c<cr><esc>')
       command('set number')
@@ -353,7 +446,7 @@ describe('Signs', function()
         {1:>>>>>>>>}{6:  1 }a                                        |
         {2:        }{6:  2 }b                                        |
         {2:        }{6:  3 }c                                        |
-        {2:        }{6:^  4 }                                         |
+        {2:        }{6:  4 }^                                         |
         {0:~                                                    }|
         {0:~                                                    }|
         {0:~                                                    }|
@@ -375,7 +468,7 @@ describe('Signs', function()
         {1:>>>>>>>>>>}{6:  1 }a                                      |
         {2:          }{6:  2 }b                                      |
         {2:          }{6:  3 }c                                      |
-        {2:        ^  }{6:  4 }                                       |
+        {2:          }{6:  4 }^                                       |
         {0:~                                                    }|
         {0:~                                                    }|
         {0:~                                                    }|
@@ -389,7 +482,7 @@ describe('Signs', function()
       ]])
     end)
 
-    it('ignores signs with no icon and text when calculting the signcolumn width', function()
+    it('ignores signs with no icon and text when calculating the signcolumn width', function()
       feed('ia<cr>b<cr>c<cr><esc>')
       command('set number')
       command('set signcolumn=auto:2')
@@ -419,7 +512,33 @@ describe('Signs', function()
         {1:>>}{6:  1 }a                                              |
         {2:  }{6:  2 }b                                              |
         {2:  }{6:  3 }c                                              |
-        {2:  }{6:  ^4 }                                               |
+        {2:  }{6:  4 }^                                               |
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+        {0:~                                                    }|
+                                                             |
+      ]])
+    end)
+
+    it('shows the line number when signcolumn=number but no marks on a line have text', function()
+      feed('ia<cr>b<cr>c<cr><esc>')
+      command('set number signcolumn=number')
+      command('sign define pietSearch text=>> texthl=Search numhl=Error')
+      command('sign define pietError text=    texthl=Search numhl=Error')
+      command('sign place 1 line=1 name=pietSearch buffer=1')
+      command('sign place 2 line=2 name=pietError  buffer=1')
+      -- no signcolumn, line number for "a" is Search, for "b" is Error, for "c" is LineNr
+      screen:expect([[
+        {1: >> }a                                                |
+        {8:  2 }b                                                |
+        {6:  3 }c                                                |
+        {6:  4 }^                                                 |
         {0:~                                                    }|
         {0:~                                                    }|
         {0:~                                                    }|
